@@ -4,12 +4,11 @@
  +-------------------------------------------------------------------------+
  | User Interface for the Enigma Plugin                                    |
  |                                                                         |
- | Copyright (C) 2010-2015 The Roundcube Dev Team                          |
+ | Copyright (C) The Roundcube Dev Team                                    |
  |                                                                         |
  | Licensed under the GNU General Public License version 3 or              |
  | any later version with exceptions for skins & plugins.                  |
  | See the README file for a full license statement.                       |
- |                                                                         |
  +-------------------------------------------------------------------------+
  | Author: Aleksander Machniak <alec@alec.pl>                              |
  +-------------------------------------------------------------------------+
@@ -59,6 +58,10 @@ class enigma_ui
                     $this->key_import();
                     break;
 
+                case 'import-search':
+                    $this->key_import_search();
+                    break;
+
                 case 'export':
                     $this->key_export();
                     break;
@@ -83,7 +86,6 @@ class enigma_ui
 
             $this->rc->output->add_handlers(array(
                     'keyslist'     => array($this, 'tpl_keys_list'),
-                    'keyframe'     => array($this, 'tpl_key_frame'),
                     'countdisplay' => array($this, 'tpl_keys_rowcount'),
                     'searchform'   => array($this->rc->output, 'search_form'),
             ));
@@ -136,6 +138,8 @@ class enigma_ui
 
         $this->enigma->include_script('enigma.js');
 
+        $this->rc->output->set_env('keyservers', $this->rc->config->keyservers());
+
         $this->js_loaded = true;
     }
 
@@ -187,18 +191,6 @@ class enigma_ui
         $this->add_js();
     }
 
-    /**
-     * Template object for key info/edit frame.
-     *
-     * @param array Object attributes
-     *
-     * @return string HTML output
-     */
-    function tpl_key_frame($attrib)
-    {
-        return $this->rc->output->frame($attrib, true);
-    }
-
     /**
      * Template object for list of keys.
      *
@@ -225,7 +217,9 @@ class enigma_ui
 
         // add some labels to client
         $this->rc->output->add_label('enigma.keyremoveconfirm', 'enigma.keyremoving',
-            'enigma.keyexportprompt', 'enigma.withprivkeys', 'enigma.onlypubkeys', 'enigma.exportkeys'
+            'enigma.keyexportprompt', 'enigma.withprivkeys', 'enigma.onlypubkeys',
+            'enigma.exportkeys', 'enigma.importkeys', 'enigma.keyimportsearchlabel',
+            'import', 'search'
         );
 
         return $out;
@@ -366,11 +360,11 @@ class enigma_ui
         $table = new html_table(array('cols' => 2));
 
         // Key user ID
-        $table->add('title', $this->enigma->gettext('keyuserid'));
+        $table->add('title', html::label(null, $this->enigma->gettext('keyuserid')));
         $table->add(null, rcube::Q($this->data->name));
 
         // Key ID
-        $table->add('title', $this->enigma->gettext('keyid'));
+        $table->add('title', html::label(null, $this->enigma->gettext('keyid')));
         $table->add(null, $this->data->subkeys[0]->get_short_id());
 
         // Key type
@@ -381,11 +375,11 @@ class enigma_ui
         else if ($keytype == enigma_key::TYPE_PUBLIC) {
             $type = $this->enigma->gettext('typepublickey');
         }
-        $table->add('title', $this->enigma->gettext('keytype'));
+        $table->add('title', html::label(null, $this->enigma->gettext('keytype')));
         $table->add(null, $type);
 
         // Key fingerprint
-        $table->add('title', $this->enigma->gettext('fingerprint'));
+        $table->add('title', html::label(null, $this->enigma->gettext('fingerprint')));
         $table->add(null, $this->data->subkeys[0]->get_fingerprint());
 
         $out .= html::tag('fieldset', null,
@@ -423,12 +417,12 @@ class enigma_ui
                 }
             }
 
+            $table->set_row_attribs($subkey->revoked || ($subkey->expires && $subkey->expires < $now) ? 'deleted' : '');
             $table->add('id', $subkey->get_short_id());
             $table->add('algo', $algo);
             $table->add('created', $subkey->created ? $this->rc->format_date($subkey->created, $date_format, false) : '');
             $table->add('expires', $subkey->expires ? $this->rc->format_date($subkey->expires, $date_format, false) : $this->enigma->gettext('expiresnever'));
             $table->add('usage', implode(',', $usage));
-            $table->set_row_attribs($subkey->revoked || ($subkey->expires && $subkey->expires < $now) ? 'deleted' : '');
         }
 
         $out .= html::tag('fieldset', null,
@@ -448,9 +442,9 @@ class enigma_ui
             }
             $username .= ' <' . $user->email . '>';
 
+            $table->set_row_attribs($user->revoked || !$user->valid ? 'deleted' : '');
             $table->add('id', rcube::Q(trim($username)));
             $table->add('valid', $this->enigma->gettext($user->valid ? 'valid' : 'unknown'));
-            $table->set_row_attribs($user->revoked || !$user->valid ? 'deleted' : '');
         }
 
         $out .= html::tag('fieldset', null,
@@ -557,12 +551,11 @@ class enigma_ui
                 if ($result['imported']) {
                     $this->rc->output->command('parent.enigma_list', 1);
                 }
-                else {
-                    $this->rc->output->command('parent.enigma_loadframe');
-                }
 
                 $this->rc->output->show_message('enigma.keysimportsuccess', 'confirmation',
                     array('new' => $result['imported'], 'old' => $result['unchanged']));
+
+                $this->rc->output->command('parent.enigma_import_success');
             }
             else if ($result instanceof enigma_error && $result->getCode() == enigma_error::BADPASS) {
                 $this->password_prompt($result);
@@ -570,7 +563,6 @@ class enigma_ui
             else {
                 $this->rc->output->show_message('enigma.keysimportfailed', 'error');
             }
-
             $this->rc->output->send('iframe');
         }
         else if ($err = $_FILES['_file']['error']) {
@@ -588,10 +580,21 @@ class enigma_ui
             'importform' => array($this, 'tpl_key_import_form'),
         ));
 
-        $this->rc->output->set_pagetitle($this->enigma->gettext('keyimport'));
         $this->rc->output->send('enigma.keyimport');
     }
 
+    /**
+     * Key import-search (page) handler
+     */
+    private function key_import_search()
+    {
+        $this->rc->output->add_handlers(array(
+            'importform' => array($this, 'tpl_key_import_form'),
+        ));
+
+        $this->rc->output->send('enigma.keysearch');
+    }
+
     /**
      * Template object for key import (upload) form
      */
@@ -599,46 +602,63 @@ class enigma_ui
     {
         $attrib += array('id' => 'rcmKeyImportForm');
 
-        $upload = new html_inputfield(array('type' => 'file', 'name' => '_file',
-            'id' => 'rcmimportfile', 'size' => 30));
-        $search = new html_inputfield(array('type' => 'text', 'name' => '_search',
-            'id' => 'rcmimportsearch', 'size' => 30));
+        if (empty($attrib['part']) || $attrib['part'] == 'import') {
+            $title  = $this->enigma->gettext('keyimportlabel');
+            $upload = new html_inputfield(array('type' => 'file', 'name' => '_file',
+                'id' => 'rcmimportfile', 'size' => 30));
 
-        $upload_button = new html_inputfield(array(
-                'type'    => 'button',
-                'value'   => $this->rc->gettext('import'),
-                'class'   => 'button',
-                'onclick' => "return rcmail.command('plugin.enigma-import','',this,event)",
-        ));
+            $max_filesize  = $this->rc->upload_init();
+            $upload_button = new html_button(array(
+                    'class'   => 'button import',
+                    'onclick' => "return rcmail.command('plugin.enigma-import','',this,event)",
+            ));
 
-        $search_button = new html_inputfield(array(
-                'type'    => 'button',
-                'value'   => $this->rc->gettext('search'),
-                'class'   => 'button',
-                'onclick' => "return rcmail.command('plugin.enigma-import-search','',this,event)",
-        ));
+            $form = html::div(null, html::p(null, rcube::Q($this->enigma->gettext('keyimporttext'), 'show'))
+                . $upload->show()
+                . html::div('hint', $this->rc->gettext(array('id' => 'importfile', 'name' => 'maxuploadsize', 'vars' => array('size' => $max_filesize))))
+                . (empty($attrib['part']) ? html::br() . html::br() . $upload_button->show($this->rc->gettext('import')) : '')
+            );
 
-        $upload_form = html::div(null,
-            rcube::Q($this->enigma->gettext('keyimporttext'), 'show')
-            . html::br() . html::br() . $upload->show()
-            . html::br() . html::br() . $upload_button->show()
-        );
+            if (empty($attrib['part'])) {
+                $form = html::tag('fieldset', '', html::tag('legend', null, $title) . $form);
+            }
+            else {
+                $this->rc->output->set_pagetitle($title);
+            }
+        }
 
-        $search_form = html::div(null,
-            rcube::Q($this->enigma->gettext('keyimportsearchtext'), 'show')
-            . html::br() . html::br() . $search->show()
-            . html::br() . html::br() . $search_button->show()
-        );
+        if (empty($attrib['part']) || $attrib['part'] == 'search') {
+            $title  = $this->enigma->gettext('keyimportsearchlabel');
+            $search = new html_inputfield(array('type' => 'text', 'name' => '_search',
+                'id' => 'rcmimportsearch', 'size' => 30));
 
-        $form = html::tag('fieldset', '', html::tag('legend', null, $this->enigma->gettext('keyimportlabel')) . $upload_form)
-            . html::tag('fieldset', '', html::tag('legend', null, $this->enigma->gettext('keyimportsearchlabel')) . $search_form);
+            $search_button = new html_button(array(
+                    'class'   => 'button search',
+                    'onclick' => "return rcmail.command('plugin.enigma-import-search','',this,event)",
+            ));
+
+            $form = html::div(null,
+                rcube::Q($this->enigma->gettext('keyimportsearchtext'), 'show')
+                . html::br() . html::br() . $search->show()
+                . (empty($attrib['part']) ? html::br() . html::br() . $search_button->show($this->rc->gettext('search')) : '')
+            );
+
+            if (empty($attrib['part'])) {
+                $form = html::tag('fieldset', '', html::tag('legend', null, $title) . $form);
+            }
+            else {
+                $this->rc->output->set_pagetitle($title);
+            }
+
+            $this->rc->output->include_script('publickey.js');
+        }
 
         $this->rc->output->add_label('selectimportfile', 'importwait', 'nopubkeyfor', 'nopubkeyforsender',
             'encryptnoattachments','encryptedsendialog','searchpubkeyservers', 'importpubkeys',
             'encryptpubkeysfound',  'search', 'close', 'import', 'keyid', 'keylength', 'keyexpired',
             'keyrevoked', 'keyimportsuccess', 'keyservererror');
+
         $this->rc->output->add_gui_object('importform', $attrib['id']);
-        $this->rc->output->include_script('publickey.js');
 
         $out = $this->rc->output->form_tag(array(
             'action'  => $this->rc->url(array('action' => $this->rc->action, 'a' => 'import')),
@@ -728,11 +748,11 @@ class enigma_ui
         foreach ($identities as $idx => $ident) {
             $name = empty($ident['name']) ? ($ident['email']) : $ident['ident'];
             $attr = array('value' => $idx, 'data-name' => $ident['name'], 'data-email' => $ident['email']);
-            $identities[$idx] = html::label(null, $checkbox->show($idx, $attr) . rcube::Q($name));
+            $identities[$idx] = html::tag('li', null, html::label(null, $checkbox->show($idx, $attr) . rcube::Q($name)));
         }
 
         $table->add('title', html::label('key-name', rcube::Q($this->enigma->gettext('newkeyident'))));
-        $table->add(null, implode($identities, "\n"));
+        $table->add(null, html::tag('ul', 'proplist', implode("\n", $identities)));
 
         // Key size
         $select = new html_select(array('name' => 'size', 'id' => 'key-size'));
@@ -744,12 +764,22 @@ class enigma_ui
 
         // Password and confirm password
         $table->add('title', html::label('key-pass', rcube::Q($this->enigma->gettext('newkeypass'))));
-        $table->add(null, rcube_output::get_edit_field('password', '',
-            array('id' => 'key-pass', 'size' => $attrib['size'], 'required' => true), 'password'));
+        $table->add(null, rcube_output::get_edit_field('password', '', array(
+                'id'           => 'key-pass',
+                'size'         => $attrib['size'],
+                'required'     => true,
+                'autocomplete' => 'new-password',
+                'oninput'      => "this.type = this.value.length ? 'password' : 'text'",
+            ), 'text'));
 
         $table->add('title', html::label('key-pass-confirm', rcube::Q($this->enigma->gettext('newkeypassconfirm'))));
-        $table->add(null, rcube_output::get_edit_field('password-confirm', '',
-            array('id' => 'key-pass-confirm', 'size' => $attrib['size'], 'required' => true), 'password'));
+        $table->add(null, rcube_output::get_edit_field('password-confirm', '', array(
+                'id'           => 'key-pass-confirm',
+                'size'         => $attrib['size'],
+                'required'     => true,
+                'autocomplete' => 'new-password',
+                'oninput'      => "this.type = this.value.length ? 'password' : 'text'",
+            ), 'text'));
 
         $this->rc->output->add_gui_object('keyform', $attrib['id']);
         $this->rc->output->add_label('enigma.keygenerating', 'enigma.formerror',
@@ -788,58 +818,81 @@ class enigma_ui
     private function compose_ui()
     {
         $this->add_css();
+        $this->rc->output->add_label('enigma.sendunencrypted');
 
-        // Options menu button
-        $this->enigma->add_button(array(
-            'type'     => 'link',
-            'command'  => 'plugin.enigma',
-            'onclick'  => "rcmail.command('menu-open', 'enigmamenu', event.target, event)",
-            'class'    => 'button enigma',
-            'title'    => 'encryptionoptions',
-            'label'    => 'encryption',
-            'domain'   => $this->enigma->ID,
-            'width'    => 32,
-            'height'   => 32,
-            'aria-owns'     => 'enigmamenu',
-            'aria-haspopup' => 'true',
-            'aria-expanded' => 'false',
-            ), 'toolbar');
+        // Elastic skin (or a skin based on it)
+        if (array_key_exists('elastic', (array) $this->rc->output->skins)) {
+            $this->enigma->api->add_content($this->compose_ui_options(), 'composeoptions');
+        }
+        // other skins
+        else {
+            // Options menu button
+            $this->enigma->add_button(array(
+                'type'     => 'link',
+                'command'  => 'plugin.enigma',
+                'onclick'  => "rcmail.command('menu-open', 'enigmamenu', event.target, event)",
+                'class'    => 'button enigma',
+                'title'    => 'encryptionoptions',
+                'label'    => 'encryption',
+                'domain'   => $this->enigma->ID,
+                'width'    => 32,
+                'height'   => 32,
+                'aria-owns'     => 'enigmamenu',
+                'aria-haspopup' => 'true',
+                'aria-expanded' => 'false',
+                ), 'toolbar');
 
+            // Options menu contents
+            $this->rc->output->add_footer($this->compose_ui_options(true));
+        }
+    }
+
+    /**
+     * Init compose UI (add task button and the menu)
+     */
+    private function compose_ui_options($wrap = false)
+    {
         $locks = (array) $this->rc->config->get('enigma_options_lock');
-        $menu  = new html_table(array('cols' => 2));
         $chbox = new html_checkbox(array('value' => 1));
 
-        $menu->add(null, html::label(array('for' => 'enigmasignopt'),
-            rcube::Q($this->enigma->gettext('signmsg'))));
-        $menu->add(null, $chbox->show($this->rc->config->get('enigma_sign_all') ? 1 : 0,
-                array(
-                    'name'     => '_enigma_sign',
-                    'id'       => 'enigmasignopt',
-                    'disabled' => in_array('sign', $locks),
-                )));
+        $out = html::div('form-group form-check row',
+            html::label(array('for' => 'enigmasignopt', 'class' => 'col-form-label col-6'),
+                rcube::Q($this->enigma->gettext('signmsg')))
+            . html::div('form-check col-6',
+                $chbox->show($this->rc->config->get('enigma_sign_all') ? 1 : 0, array(
+                        'name'     => '_enigma_sign',
+                        'id'       => 'enigmasignopt',
+                        'class'    => 'form-check-input',
+                        'disabled' => in_array('sign', $locks),
+                ))));
 
-        $menu->add(null, html::label(array('for' => 'enigmaencryptopt'),
-            rcube::Q($this->enigma->gettext('encryptmsg'))));
-        $menu->add(null, $chbox->show($this->rc->config->get('enigma_encrypt_all') ? 1 : 0,
-                array(
-                    'name'     => '_enigma_encrypt',
-                    'id'       => 'enigmaencryptopt',
-                    'disabled' => in_array('encrypt', $locks),
-                )));
+        $out .= html::div('form-group form-check row',
+            html::label(array('for' => 'enigmaencryptopt', 'class' => 'col-form-label col-6'),
+                rcube::Q($this->enigma->gettext('encryptmsg')))
+            . html::div('form-check col-6',
+                $chbox->show($this->rc->config->get('enigma_encrypt_all') ? 1 : 0, array(
+                        'name'     => '_enigma_encrypt',
+                        'id'       => 'enigmaencryptopt',
+                        'class'    => 'form-check-input',
+                        'disabled' => in_array('encrypt', $locks),
+                ))));
 
-        $menu->add(null, html::label(array('for' => 'enigmaattachpubkeyopt'),
-            rcube::Q($this->enigma->gettext('attachpubkeymsg'))));
-        $menu->add(null, $chbox->show($this->rc->config->get('enigma_attach_pubkey') ? 1 : 0,
-                array(
-                    'name'     => '_enigma_attachpubkey',
-                    'id'       => 'enigmaattachpubkeyopt',
-                    'disabled' => in_array('pubkey', $locks),
-                )));
+        $out .= html::div('form-group form-check row',
+            html::label(array('for' => 'enigmaattachpubkeyopt', 'class' => 'col-form-label col-6'),
+                rcube::Q($this->enigma->gettext('attachpubkeymsg')))
+            . html::div('form-check col-6',
+                $chbox->show($this->rc->config->get('enigma_attach_pubkey') ? 1 : 0, array(
+                        'name'     => '_enigma_attachpubkey',
+                        'id'       => 'enigmaattachpubkeyopt',
+                        'class'    => 'form-check-input',
+                        'disabled' => in_array('pubkey', $locks),
+                ))));
 
-        $menu = html::div(array('id' => 'enigmamenu', 'class' => 'popupmenu'), $menu->show());
+        if (!$wrap) {
+            return $out;
+        }
 
-        // Options menu contents
-        $this->rc->output->add_footer($menu);
+        return html::div(array('id' => 'enigmamenu', 'class' => 'popupmenu'), $out);
     }
 
     /**
@@ -881,7 +934,7 @@ class enigma_ui
             $attrib['id'] = 'enigma-message';
 
             if ($status instanceof enigma_error) {
-                $attrib['class'] = 'enigmaerror';
+                $attrib['class'] = 'boxerror enigmaerror encrypted';
                 $code            = $status->getCode();
 
                 if ($code == enigma_error::KEYNOTFOUND) {
@@ -902,11 +955,11 @@ class enigma_ui
                 }
             }
             else if ($status === enigma_engine::ENCRYPTED_PARTIALLY) {
-                $attrib['class'] = 'enigmawarning';
+                $attrib['class'] = 'boxwarning enigmawarning encrypted';
                 $msg = rcube::Q($this->enigma->gettext('decryptpartial'));
             }
             else {
-                $attrib['class'] = 'enigmanotice';
+                $attrib['class'] = 'boxconfirmation enigmanotice encrypted';
                 $msg = rcube::Q($this->enigma->gettext('decryptok'));
             }
 
@@ -927,24 +980,21 @@ class enigma_ui
             $attrib['id'] = 'enigma-message';
 
             if ($sig instanceof enigma_signature) {
-                $sender = $sig->name ?: '';
-                if ($sig->email) {
-                    $sender .= ' <' . $sig->email . '>';
-                }
+                $sender = $sig->get_sender($engine, $p['message'], $part_id);
 
                 if ($sig->valid === enigma_error::UNVERIFIED) {
-                    $attrib['class'] = 'enigmawarning';
+                    $attrib['class'] = 'boxwarning enigmawarning signed';
                     $msg = str_replace('$sender', $sender, $this->enigma->gettext('sigunverified'));
                     $msg = str_replace('$keyid', $sig->id, $msg);
                     $msg = rcube::Q($msg);
                 }
                 else if ($sig->valid) {
-                    $attrib['class'] = $sig->partial ? 'enigmawarning' : 'enigmanotice';
+                    $attrib['class'] = ($sig->partial ? 'boxwarning enigmawarning' : 'boxconfirmation enigmanotice') . ' signed';
                     $label = 'sigvalid' . ($sig->partial ? 'partial' : '');
                     $msg = rcube::Q(str_replace('$sender', $sender, $this->enigma->gettext($label)));
                 }
                 else {
-                    $attrib['class'] = 'enigmawarning';
+                    $attrib['class'] = 'boxwarning enigmawarning signed';
                     if ($sender) {
                         $msg = rcube::Q(str_replace('$sender', $sender, $this->enigma->gettext('siginvalid')));
                     }
@@ -955,12 +1005,12 @@ class enigma_ui
                 }
             }
             else if ($sig && $sig->getCode() == enigma_error::KEYNOTFOUND) {
-                $attrib['class'] = 'enigmawarning';
+                $attrib['class'] = 'boxwarning enigmawarning signed';
                 $msg = rcube::Q(str_replace('$keyid', enigma_key::format_id($sig->getData('id')),
                     $this->enigma->gettext('signokey')));
             }
             else {
-                $attrib['class'] = 'enigmaerror';
+                $attrib['class'] = 'boxwarning enigmaerror signed';
                 $msg = rcube::Q($this->enigma->gettext('sigerror'));
             }
 /*
@@ -1040,13 +1090,15 @@ class enigma_ui
                 $p['content'] = '';
             }
 
-            // add box below message body
-            $p['content'] .= html::p(array('class' => 'enigmaattachment'),
-                html::a(array(
-                    'href'    => "#",
-                    'onclick' => "return ".rcmail_output::JS_OBJECT_NAME.".enigma_import_attachment('".rcube::JQ($part)."')",
-                    'title'   => $this->enigma->gettext('keyattimport')),
-                    html::span(null, $this->enigma->gettext('keyattfound'))));
+            // add box above the message body
+            $p['content'] = html::p(array('class' => 'enigmaattachment boxinformation aligned-buttons'),
+                html::span(null, rcube::Q($this->enigma->gettext('keyattfound'))) .
+                html::tag('button', array(
+                        'onclick' => "return ".rcmail_output::JS_OBJECT_NAME.".enigma_import_attachment('".rcube::JQ($part)."')",
+                        'title'   => $this->enigma->gettext('keyattimport'),
+                        'class'   => 'import',
+                    ), rcube::Q($this->rc->gettext('import')))
+            ) . $p['content'];
 
             $attach_scripts = true;
         }
@@ -1065,6 +1117,11 @@ class enigma_ui
      */
     function message_ready($p)
     {
+        // The message might have been already encrypted by Mailvelope
+        if (strpos($p['message']->getParam('ctype'), 'multipart/encrypted') === 0) {
+            return $p;
+        }
+
         $savedraft      = !empty($_POST['_draft']) && empty($_GET['_saveonly']);
         $sign_enable    = (bool) rcube_utils::get_input_value('_enigma_sign', rcube_utils::INPUT_POST);
         $encrypt_enable = (bool) rcube_utils::get_input_value('_enigma_encrypt', rcube_utils::INPUT_POST);
@@ -1100,10 +1157,14 @@ class enigma_ui
 
         if ($mode && ($status instanceof enigma_error)) {
             $code = $status->getCode();
-
             if ($code == enigma_error::KEYNOTFOUND) {
-                $vars = array('email' => $status->getData('missing'));
-                $msg  = 'enigma.' . $mode . 'nokey';
+                if ($email = $status->getData('missing')) {
+                    $vars = array('email' => $email);
+                    $msg  = 'enigma.' . $mode . 'nokey';
+                }
+                else {
+                    $msg = 'enigma.' . ($encrypt_enable ? 'encryptnoprivkey' : 'signnokey');
+                }
             }
             else if ($code == enigma_error::BADPASS) {
                 $this->password_prompt($status);
@@ -1119,6 +1180,7 @@ class enigma_ui
                             'text'   => $this->rc->gettext(array('name' => $msg, 'vars' => $vars)),
                             'title'  => $this->enigma->gettext('keynotfound'),
                             'button' => $this->enigma->gettext('findkey'),
+                            'mode'   => $mode,
                     ));
                 }
                 else {
@@ -1191,7 +1253,6 @@ class enigma_ui
         $uid     = rcube_utils::get_input_value('_uid', rcube_utils::INPUT_POST);
         $mbox    = rcube_utils::get_input_value('_mbox', rcube_utils::INPUT_POST);
         $mime_id = rcube_utils::get_input_value('_part', rcube_utils::INPUT_POST);
-        $storage = $this->rc->get_storage();
         $engine  = $this->enigma->load_engine();
 
         if ($uid && $mime_id) {
