@@ -3,8 +3,9 @@
 /**
  +-----------------------------------------------------------------------+
  | This file is part of the Roundcube Webmail client                     |
- | Copyright (C) 2008-2014, The Roundcube Dev Team                       |
- | Copyright (C) 2011-2014, Kolab Systems AG                             |
+ |                                                                       |
+ | Copyright (C) The Roundcube Dev Team                                  |
+ | Copyright (C) Kolab Systems AG                                        |
  |                                                                       |
  | Licensed under the GNU General Public License version 3 or            |
  | any later version with exceptions for skins & plugins.                |
@@ -38,94 +39,60 @@ class rcube
 
     const DEBUG_LINE_LENGTH = 4096;
 
-    /**
-     * Singleton instance of rcube
-     *
-     * @var rcube
-     */
-    static protected $instance;
-
-    /**
-     * Stores instance of rcube_config.
-     *
-     * @var rcube_config
-     */
+    /** @var rcube_config Stores instance of rcube_config */
     public $config;
 
-    /**
-     * Instance of database class.
-     *
-     * @var rcube_db
-     */
+    /** @var rcube_db Instance of database class */
     public $db;
 
-    /**
-     * Instance of Memcache class.
-     *
-     * @var Memcache
-     */
+    /** @var Memcache Instance of Memcache class */
     public $memcache;
 
-   /**
-     * Instance of rcube_session class.
-     *
-     * @var rcube_session
-     */
+    /** @var Memcached Instance of Memcached class */
+    public $memcached;
+
+    /** @var Redis Instance of Redis class */
+    public $redis;
+
+    /** @var rcube_session Instance of rcube_session class */
     public $session;
 
-    /**
-     * Instance of rcube_smtp class.
-     *
-     * @var rcube_smtp
-     */
+    /** @var rcube_smtp Instance of rcube_smtp class */
     public $smtp;
 
-    /**
-     * Instance of rcube_storage class.
-     *
-     * @var rcube_storage
-     */
+    /** @var rcube_storage Instance of rcube_storage class */
     public $storage;
 
-    /**
-     * Instance of rcube_output class.
-     *
-     * @var rcube_output
-     */
+    /** @var rcube_output Instance of rcube_output class */
     public $output;
 
-    /**
-     * Instance of rcube_plugin_api.
-     *
-     * @var rcube_plugin_api
-     */
+    /** @var rcube_plugin_api Instance of rcube_plugin_api */
     public $plugins;
 
-    /**
-     * Instance of rcube_user class.
-     *
-     * @var rcube_user
-     */
+    /** @var rcube_user Instance of rcube_user class */
     public $user;
 
-    /**
-     * Request status
-     *
-     * @var int
-     */
+    /** @var int Request status */
     public $request_status = 0;
 
-    /* private/protected vars */
+    /** @var array Localization */
     protected $texts;
+
+    /** @var rcube_cache[] Initialized cache objects */
     protected $caches = array();
+
+    /** @var array Registered shutdown functions */
     protected $shutdown_functions = array();
 
+    /** @var rcube Singleton instance of rcube */
+    static protected $instance;
+
 
     /**
      * This implements the 'singleton' design pattern
      *
-     * @param integer $mode Options to initialize with this instance. See rcube::INIT_WITH_* constants
-     * @param string  $env  Environment name to run (e.g. live, dev, test)
+     * @param int    $mode Options to initialize with this instance. See rcube::INIT_WITH_* constants
+     * @param string $env  Environment name to run (e.g. live, dev, test)
      *
      * @return rcube The one and only instance
      */
@@ -141,6 +108,8 @@ class rcube
 
     /**
      * Private constructor
+     *
+     * @param string $env Environment name to run (e.g. live, dev, test)
      */
     protected function __construct($env = '')
     {
@@ -153,6 +122,8 @@ class rcube
 
     /**
      * Initial startup function
+     *
+     * @param int $mode Options to initialize with this instance. See rcube::INIT_WITH_* constants
      */
     protected function init($mode = 0)
     {
@@ -202,86 +173,54 @@ class rcube
     public function get_memcache()
     {
         if (!isset($this->memcache)) {
-            // no memcache support in PHP
-            if (!class_exists('Memcache')) {
-                $this->memcache = false;
-                return false;
-            }
-
-            $this->memcache     = new Memcache;
-            $this->memcache_init();
-
-            // test connection and failover (will result in $this->mc_available == 0 on complete failure)
-            $this->memcache->increment('__CONNECTIONTEST__', 1);  // NOP if key doesn't exist
-
-            if (!$this->mc_available) {
-                $this->memcache = false;
-            }
+            $this->memcache = rcube_cache_memcache::engine();
         }
 
         return $this->memcache;
     }
 
     /**
-     * Get global handle for memcache access
+     * Get global handle for memcached access
      *
-     * @return object Memcache
+     * @return object Memcached
      */
-    protected function memcache_init()
+    public function get_memcached()
     {
-        $this->mc_available = 0;
-
-        // add all configured hosts to pool
-        $pconnect       = $this->config->get('memcache_pconnect', true);
-        $timeout        = $this->config->get('memcache_timeout', 1);
-        $retry_interval = $this->config->get('memcache_retry_interval', 15);
-
-        foreach ($this->config->get('memcache_hosts', array()) as $host) {
-            if (substr($host, 0, 7) != 'unix://') {
-                list($host, $port) = explode(':', $host);
-                if (!$port) $port = 11211;
-            }
-            else {
-                $port = 0;
-            }
-
-            $this->mc_available += intval($this->memcache->addServer(
-                $host, $port, $pconnect, 1, $timeout, $retry_interval, false, array($this, 'memcache_failure')));
+        if (!isset($this->memcached)) {
+            $this->memcached = rcube_cache_memcached::engine();
         }
+
+        return $this->memcached;
     }
 
     /**
-     * Callback for memcache failure
+     * Get global handle for redis access
+     *
+     * @return object Redis
      */
-    public function memcache_failure($host, $port)
+    public function get_redis()
     {
-        static $seen = array();
-
-        // only report once
-        if (!$seen["$host:$port"]++) {
-            $this->mc_available--;
-            self::raise_error(array(
-                'code' => 604, 'type' => 'db',
-                'line' => __LINE__, 'file' => __FILE__,
-                'message' => "Memcache failure on host $host:$port"),
-                true, false);
+        if (!isset($this->redis)) {
+            $this->redis = rcube_cache_redis::engine();
         }
+
+        return $this->redis;
     }
 
     /**
-     * Initialize and get cache object
+     * Initialize and get user cache object
      *
      * @param string $name   Cache identifier
-     * @param string $type   Cache type ('db', 'apc' or 'memcache')
+     * @param string $type   Cache type ('db', 'apc', 'memcache', 'redis')
      * @param string $ttl    Expiration time for cache items
      * @param bool   $packed Enables/disables data serialization
      *
-     * @return rcube_cache Cache object
+     * @return rcube_cache User cache object
      */
     public function get_cache($name, $type='db', $ttl=0, $packed=true)
     {
         if (!isset($this->caches[$name]) && ($userid = $this->get_user_id())) {
-            $this->caches[$name] = new rcube_cache($type, $userid, $name, $ttl, $packed);
+            $this->caches[$name] = rcube_cache::factory($type, $userid, $name, $ttl, $packed);
         }
 
         return $this->caches[$name];
@@ -293,9 +232,9 @@ class rcube
      * @param string $name   Cache identifier
      * @param bool   $packed Enables/disables data serialization
      *
-     * @return rcube_cache_shared Cache object
+     * @return rcube_cache Shared cache object
      */
-    public function get_cache_shared($name, $packed=true)
+    public function get_cache_shared($name, $packed = true)
     {
         $shared_name = "shared_$name";
 
@@ -313,7 +252,7 @@ class rcube
                 $ttl = $this->config->get('shared_cache_ttl', '10d');
             }
 
-            $this->caches[$shared_name] = new rcube_cache_shared($type, $name, $ttl, $packed);
+            $this->caches[$shared_name] = rcube_cache::factory($type, null, $name, $ttl, $packed);
         }
 
         return $this->caches[$shared_name];
@@ -478,6 +417,8 @@ class rcube
 
     /**
      * Callback for IMAP connection events to log session identifiers
+     *
+     * @param array $args Callback arguments
      */
     public function storage_log_session($args)
     {
@@ -491,11 +432,12 @@ class rcube
      */
     public function session_init()
     {
-        // session started (Installer?)
-        if (session_id()) {
+        // Ignore in CLI mode or when session started (Installer?)
+        if (empty($_SERVER['REMOTE_ADDR']) || session_id()) {
             return;
         }
 
+        $storage     = $this->config->get('session_storage', 'db');
         $sess_name   = $this->config->get('session_name');
         $sess_domain = $this->config->get('session_domain');
         $sess_path   = $this->config->get('session_path');
@@ -523,14 +465,16 @@ class rcube
         ini_set('session.use_only_cookies', 1);
         ini_set('session.cookie_httponly', 1);
 
-        // get session driver instance
+        // Make sure session garbage collector is enabled when using custom handlers (#6560)
+        // Note: Use session.gc_divisor to control accuracy
+        if ($storage != 'php' && !ini_get('session.gc_probability')) {
+            ini_set('session.gc_probability', 1);
+        }
+
+        // Start the session
         $this->session = rcube_session::factory($this->config);
         $this->session->register_gc_handler(array($this, 'gc'));
-
-        // start PHP session (if not in CLI mode)
-        if ($_SERVER['REMOTE_ADDR']) {
-            $this->session->start();
-        }
+        $this->session->start();
     }
 
     /**
@@ -539,15 +483,13 @@ class rcube
     public function gc()
     {
         rcube_cache::gc();
-        rcube_cache_shared::gc();
         $this->get_storage()->cache_gc();
-
         $this->gc_temp();
     }
 
     /**
      * Garbage collector function for temp files.
-     * Remove temp files older than two days
+     * Removes temporary files older than temp_dir_ttl.
      */
     public function gc_temp()
     {
@@ -556,19 +498,20 @@ class rcube
         // expire in 48 hours by default
         $temp_dir_ttl = $this->config->get('temp_dir_ttl', '48h');
         $temp_dir_ttl = get_offset_sec($temp_dir_ttl);
-        if ($temp_dir_ttl < 6*3600)
+        if ($temp_dir_ttl < 6*3600) {
             $temp_dir_ttl = 6*3600;   // 6 hours sensible lower bound.
+        }
 
         $expire = time() - $temp_dir_ttl;
 
         if ($tmp && ($dir = opendir($tmp))) {
             while (($fname = readdir($dir)) !== false) {
-                if ($fname[0] == '.') {
+                if (strpos($fname, RCUBE_TEMP_FILE_PREFIX) !== 0) {
                     continue;
                 }
 
-                if (@filemtime($tmp.'/'.$fname) < $expire) {
-                    @unlink($tmp.'/'.$fname);
+                if (@filemtime("$tmp/$fname") < $expire) {
+                    @unlink("$tmp/$fname");
                 }
             }
 
@@ -660,9 +603,9 @@ class rcube
     /**
      * Check if the given text label exists
      *
-     * @param string $name       Label name
-     * @param string $domain     Label domain (plugin) name or '*' for all domains
-     * @param string $ref_domain Sets domain name if label is found
+     * @param string $name        Label name
+     * @param string $domain      Label domain (plugin) name or '*' for all domains
+     * @param string &$ref_domain Sets domain name if label is found
      *
      * @return boolean True if text exists (either in the current language or in en_US)
      */
@@ -688,9 +631,9 @@ class rcube
             }
         }
         // specified domain
-        else if ($domain) {
+        else if ($domain && isset($this->texts[$domain.'.'.$name])) {
             $ref_domain = $domain;
-            return isset($this->texts[$domain.'.'.$name]);
+            return true;
         }
 
         return false;
@@ -750,6 +693,65 @@ class rcube
         }
     }
 
+    /**
+     * Read localized texts from an additional location (plugins, skins).
+     * Then you can use the result as 2nd arg to load_language().
+     *
+     * @param string $dir Directory to search in
+     *
+     * @return array Localization texts
+     */
+    public function read_localization($dir)
+    {
+        $lang   = $_SESSION['language'];
+        $langs  = array_unique(array('en_US', $lang));
+        $locdir = slashify($dir);
+        $texts  = array();
+
+        // Language aliases used to find localization in similar lang, see below
+        $aliases = array(
+            'de_CH' => 'de_DE',
+            'es_AR' => 'es_ES',
+            'fa_AF' => 'fa_IR',
+            'nl_BE' => 'nl_NL',
+            'pt_BR' => 'pt_PT',
+            'zh_CN' => 'zh_TW',
+        );
+
+        // use buffering to handle empty lines/spaces after closing PHP tag
+        ob_start();
+
+        foreach ($langs as $lng) {
+            $fpath = $locdir . $lng . '.inc';
+            if (is_file($fpath) && is_readable($fpath)) {
+                include $fpath;
+                $texts = (array) $labels + (array) $messages + (array) $texts;
+            }
+            else if ($lng != 'en_US') {
+                // Find localization in similar language (#1488401)
+                $alias = null;
+                if (!empty($aliases[$lng])) {
+                    $alias = $aliases[$lng];
+                }
+                else if ($key = array_search($lng, $aliases)) {
+                    $alias = $key;
+                }
+
+                if (!empty($alias)) {
+                    $fpath = $locdir . $alias . '.inc';
+                    if (is_file($fpath) && is_readable($fpath)) {
+                        include $fpath;
+                        $texts = (array) $labels + (array) $messages + (array) $texts;
+                    }
+                }
+            }
+        }
+
+        ob_end_clean();
+
+        return $texts;
+    }
+
     /**
      * Check the given string and return a valid language code
      *
@@ -832,9 +834,9 @@ class rcube
     /**
      * Encrypt a string
      *
-     * @param string  $clear  Clear text input
-     * @param string  $key    Encryption key to retrieve from the configuration, defaults to 'des_key'
-     * @param boolean $base64 Whether or not to base64_encode() the result before returning
+     * @param string $clear  Clear text input
+     * @param string $key    Encryption key to retrieve from the configuration, defaults to 'des_key'
+     * @param bool   $base64 Whether or not to base64_encode() the result before returning
      *
      * @return string Encrypted text
      */
@@ -848,7 +850,19 @@ class rcube
         $method = $this->config->get_crypto_method();
         $opts   = defined('OPENSSL_RAW_DATA') ? OPENSSL_RAW_DATA : true;
         $iv     = rcube_utils::random_bytes(openssl_cipher_iv_length($method), true);
-        $cipher = $iv . openssl_encrypt($clear, $method, $ckey, $opts, $iv);
+        $cipher = openssl_encrypt($clear, $method, $ckey, $opts, $iv);
+
+        if ($cipher === false) {
+            self::raise_error(array(
+                    'file'    => __FILE__,
+                    'line'    => __LINE__,
+                    'message' => "Failed to encrypt data with configured cipher method: $method!"
+                ), true, false);
+
+            return false;
+        }
+
+        $cipher = $iv . $cipher;
 
         return $base64 ? base64_encode($cipher) : $cipher;
     }
@@ -856,9 +870,9 @@ class rcube
     /**
      * Decrypt a string
      *
-     * @param string  $cipher Encrypted text
-     * @param string  $key    Encryption key to retrieve from the configuration, defaults to 'des_key'
-     * @param boolean $base64 Whether or not input is base64-encoded
+     * @param string $cipher Encrypted text
+     * @param string $key    Encryption key to retrieve from the configuration, defaults to 'des_key'
+     * @param boo    $base64 Whether or not input is base64-encoded
      *
      * @return string Decrypted text
      */
@@ -936,7 +950,7 @@ class rcube
      *
      * @param int $mode Request method
      *
-     * @return boolean True if request token is valid false if not
+     * @return bool True if request token is valid false if not
      */
     public function check_request($mode = rcube_utils::INPUT_POST)
     {
@@ -1043,7 +1057,7 @@ class rcube
 
     /**
      * When you're going to sleep the script execution for a longer time
-     * it is good to close all external connections (sql, memcache, SMTP, IMAP).
+     * it is good to close all external connections (sql, memcache, redis, SMTP, IMAP).
      *
      * No action is required on wake up, all connections will be
      * re-established automatically.
@@ -1066,19 +1080,29 @@ class rcube
 
         if ($this->memcache) {
             $this->memcache->close();
-            // after close() need to re-init memcache
-            $this->memcache_init();
+        }
+
+        if ($this->memcached) {
+            $this->memcached->quit();
         }
 
         if ($this->smtp) {
             $this->smtp->disconnect();
         }
+
+        if ($this->redis) {
+            $this->redis->close();
+        }
     }
 
     /**
      * Quote a given string.
      * Shortcut function for rcube_utils::rep_specialchars_output()
      *
+     * @param string $str      A string to quote
+     * @param string $mode     Replace mode for tags: show|remove|strict
+     * @param bool   $newlines Convert newlines
+     *
      * @return string HTML-quoted string
      */
     public static function Q($str, $mode = 'strict', $newlines = true)
@@ -1090,6 +1114,8 @@ class rcube
      * Quote a given string for javascript output.
      * Shortcut function for rcube_utils::rep_specialchars_output()
      *
+     * @param string $str A string to quote
+     *
      * @return string JS-quoted string
      */
     public static function JQ($str)
@@ -1101,10 +1127,10 @@ class rcube
      * Construct shell command, execute it and return output as string.
      * Keywords {keyword} are replaced with arguments
      *
-     * @param $cmd    Format string with {keywords} to be replaced
-     * @param $values (zero, one or more arrays can be passed)
+     * @param string $cmd        Format string with {keywords} to be replaced
+     * @param mixed  $values,... (zero, one or more arrays can be passed)
      *
-     * @return output of command. shell errors not detectable
+     * @return string Output of command. Shell errors not detectable
      */
     public static function exec(/* $cmd, $values1 = array(), ... */)
     {
@@ -1140,13 +1166,13 @@ class rcube
                 }
             }
 
-            $replacements[$tag] = join(" ", $parts);
+            $replacements[$tag] = implode(' ', $parts);
         }
 
         // use strtr behaviour of going through source string once
         $cmd = strtr($cmd, $replacements);
 
-        return (string)shell_exec($cmd);
+        return (string) shell_exec($cmd);
     }
 
     /**
@@ -1173,7 +1199,7 @@ class rcube
             $msg[] = !is_string($arg) ? var_export($arg, true) : $arg;
         }
 
-        self::write_log('console', join(";\n", $msg));
+        self::write_log('console', implode(";\n", $msg));
     }
 
     /**
@@ -1236,9 +1262,12 @@ class rcube
         $line = sprintf("[%s]: %s\n", $date, $line);
 
         // per-user logging is activated
-        if (self::$instance && self::$instance->config->get('per_user_logging') && self::$instance->get_user_id()) {
+        if (self::$instance && self::$instance->config->get('per_user_logging')
+            && self::$instance->get_user_id()
+            && !in_array($name, array('userlogins', 'sendmail'))
+        ) {
             $log_dir = self::$instance->get_user_log_dir();
-            if (empty($log_dir) && !in_array($name, array('errors', 'userlogins', 'sendmail'))) {
+            if (empty($log_dir) && $name !== 'errors') {
                 return false;
             }
         }
@@ -1256,6 +1285,13 @@ class rcube
             $log_dir = RCUBE_INSTALL_PATH . 'logs';
         }
 
+        if (self::$instance) {
+            $name .= self::$instance->config->get('log_file_ext', '.log');
+        }
+        else {
+            $name .= '.log';
+        }
+
         return file_put_contents("$log_dir/$name", $line, FILE_APPEND) !== false;
     }
 
@@ -1268,13 +1304,13 @@ class rcube
      *      - message: Error message
      *      - file:    File where error occurred
      *      - line:    Line where error occurred
-     * @param boolean $log       True to log the error
-     * @param boolean $terminate Terminate script execution
+     * @param bool $log       True to log the error
+     * @param bool $terminate Terminate script execution
      */
     public static function raise_error($arg = array(), $log = false, $terminate = false)
     {
         // handle PHP exceptions
-        if (is_object($arg) && is_a($arg, 'Exception')) {
+        if ($arg instanceof Exception) {
             $arg = array(
                 'code' => $arg->getCode(),
                 'line' => $arg->getLine(),
@@ -1282,6 +1318,13 @@ class rcube
                 'message' => $arg->getMessage(),
             );
         }
+        else if ($arg instanceof PEAR_Error) {
+            $info = $arg->getUserInfo();
+            $arg  = array(
+                'code'    => $arg->getCode(),
+                'message' => $arg->getMessage() . ($info ? ': ' . $info : ''),
+            );
+        }
         else if (is_string($arg)) {
             $arg = array('message' => $arg);
         }
@@ -1292,6 +1335,13 @@ class rcube
 
         $cli = php_sapi_name() == 'cli';
 
+        $arg['cli'] = $cli;
+        $arg['log'] = $log;
+        $arg['terminate'] = $terminate;
+
+        // send error to external error tracking tool
+        $arg = self::$instance->plugins->exec_hook('raise_error', $arg);
+
         // installer
         if (!$cli && class_exists('rcmail_install', false)) {
             $rci = rcmail_install::get_instance();
@@ -1304,25 +1354,24 @@ class rcube
             self::log_bug($arg);
         }
 
+        if ($cli) {
+            fwrite(STDERR, 'ERROR: ' . trim($arg['message']) . "\n");
+        }
+        else if ($terminate && is_object(self::$instance->output)) {
+            self::$instance->output->raise_error($arg['code'], $arg['message']);
+        }
+        else if ($terminate) {
+            header("HTTP/1.0 500 Internal Error");
+        }
+
         // terminate script
         if ($terminate) {
-            // display error page
-            if (is_object(self::$instance->output)) {
-                self::$instance->output->raise_error($arg['code'], $arg['message']);
-            }
-            else if ($cli) {
-                fwrite(STDERR, 'ERROR: ' . $arg['message']);
-            }
-
             exit(1);
         }
-        else if ($cli) {
-            fwrite(STDERR, 'ERROR: ' . $arg['message']);
-        }
     }
 
     /**
-     * Report error according to configured debug_level
+     * Log an error
      *
      * @param array $arg_arr Named parameters
      * @see self::raise_error()
@@ -1330,65 +1379,38 @@ class rcube
     public static function log_bug($arg_arr)
     {
         $program = strtoupper($arg_arr['type'] ?: 'php');
-        $level   = self::get_instance()->config->get('debug_level');
-
-        // disable errors for ajax requests, write to log instead (#1487831)
-        if (($level & 4) && !empty($_REQUEST['_remote'])) {
-            $level = ($level ^ 4) | 1;
-        }
 
         // write error to local log file
-        if (($level & 1) || !empty($arg_arr['fatal'])) {
-            if ($_SERVER['REQUEST_METHOD'] == 'POST') {
-                foreach (array('_task', '_action') as $arg) {
-                    if ($_POST[$arg] && !$_GET[$arg]) {
-                        $post_query[$arg] = $_POST[$arg];
-                    }
-                }
-
-                if (!empty($post_query)) {
-                    $post_query = (strpos($_SERVER['REQUEST_URI'], '?') != false ? '&' : '?')
-                        . http_build_query($post_query, '', '&');
+        if ($_SERVER['REQUEST_METHOD'] == 'POST') {
+            foreach (array('_task', '_action') as $arg) {
+                if ($_POST[$arg] && !$_GET[$arg]) {
+                    $post_query[$arg] = $_POST[$arg];
                 }
             }
 
-            $log_entry = sprintf("%s Error: %s%s (%s %s)",
-                $program,
-                $arg_arr['message'],
-                $arg_arr['file'] ? sprintf(' in %s on line %d', $arg_arr['file'], $arg_arr['line']) : '',
-                $_SERVER['REQUEST_METHOD'],
-                $_SERVER['REQUEST_URI'] . $post_query);
-
-            if (!self::write_log('errors', $log_entry)) {
-                // send error to PHPs error handler if write_log didn't succeed
-                trigger_error($arg_arr['message'], E_USER_WARNING);
+            if (!empty($post_query)) {
+                $post_query = (strpos($_SERVER['REQUEST_URI'], '?') != false ? '&' : '?')
+                    . http_build_query($post_query, '', '&');
             }
         }
 
-        // report the bug to the global bug reporting system
-        if ($level & 2) {
-            // TODO: Send error via HTTP
-        }
+        $log_entry = sprintf("%s Error: %s%s (%s %s)",
+            $program,
+            $arg_arr['message'],
+            $arg_arr['file'] ? sprintf(' in %s on line %d', $arg_arr['file'], $arg_arr['line']) : '',
+            $_SERVER['REQUEST_METHOD'],
+            $_SERVER['REQUEST_URI'] . $post_query);
 
-        // show error if debug_mode is on
-        if ($level & 4) {
-            print "<b>$program Error";
-
-            if (!empty($arg_arr['file']) && !empty($arg_arr['line'])) {
-                print " in $arg_arr[file] ($arg_arr[line])";
-            }
-
-            print ':</b>&nbsp;';
-            print nl2br($arg_arr['message']);
-            print '<br />';
-            flush();
+        if (!self::write_log('errors', $log_entry)) {
+            // send error to PHPs error handler if write_log didn't succeed
+            trigger_error($arg_arr['message'], E_USER_WARNING);
         }
     }
 
     /**
      * Write debug info to the log
      *
-     * @param string $engine Engine type - file name (memcache, apc)
+     * @param string $engine Engine type - file name (memcache, apc, redis)
      * @param string $data   Data string to log
      * @param bool   $result Operation result
      */
@@ -1472,8 +1494,6 @@ class rcube
         else if (isset($_SESSION['user_id'])) {
             return $_SESSION['user_id'];
         }
-
-        return null;
     }
 
     /**
@@ -1520,6 +1540,8 @@ class rcube
 
     /**
      * Get the per-user log directory
+     *
+     * @return string|false Per-user log directory if it exists and is writable, False otherwise
      */
     protected function get_user_log_dir()
     {
@@ -1581,16 +1603,17 @@ class rcube
     /**
      * Send the given message using the configured method.
      *
-     * @param object $message   Reference to Mail_MIME object
-     * @param string $from      Sender address string
-     * @param array  $mailto    Array of recipient address strings
-     * @param array  $error     SMTP error array (reference)
-     * @param string $body_file Location of file with saved message body (reference),
-     *                          used when delay_file_io is enabled
-     * @param array  $options   SMTP options (e.g. DSN request)
-     * @param bool   $disconnect Close SMTP connection ASAP
+     * @param Mail_Mime    &$message    Reference to Mail_MIME object
+     * @param string       $from        Sender address string
+     * @param array|string $mailto      Either a comma-separated list of recipients (RFC822 compliant),
+     *                                  or an array of recipients, each RFC822 valid
+     * @param array        &$error      SMTP error array (reference)
+     * @param string       &$body_file  Location of file with saved message body (reference),
+     *                                  used when delay_file_io is enabled
+     * @param array        $options     SMTP options (e.g. DSN request)
+     * @param bool         $disconnect  Close SMTP connection ASAP
      *
-     * @return boolean Send status.
+     * @return bool Send status.
      */
     public function deliver_message(&$message, $from, $mailto, &$error,
         &$body_file = null, $options = null, $disconnect = false)
@@ -1634,8 +1657,7 @@ class rcube
 
         if ($message->getParam('delay_file_io')) {
             // use common temp dir
-            $temp_dir    = $this->config->get('temp_dir');
-            $body_file   = tempnam($temp_dir, 'rcmMsg');
+            $body_file   = rcube_utils::temp_filename('msg');
             $mime_result = $message->saveMessageBody($body_file);
 
             if (is_a($mime_result, 'PEAR_Error')) {
@@ -1665,7 +1687,7 @@ class rcube
         if (!$sent) {
             self::raise_error(array('code' => 800, 'type' => 'smtp',
                 'line' => __LINE__, 'file' => __FILE__,
-                'message' => join("\n", $response)), true, false);
+                'message' => implode("\n", $response)), true, false);
 
             // allow plugins to catch sending errors with the same parameters as in 'message_before_send'
             $this->plugins->exec_hook('message_send_error', $plugin + array('error' => $error));
@@ -1681,11 +1703,12 @@ class rcube
                 $mailto = implode(',', $a_recipients);
                 $mailto = rcube_mime::decode_address_list($mailto, null, false, null, true);
 
-                self::write_log('sendmail', sprintf("User %s [%s]; Message for %s; %s",
+                self::write_log('sendmail', sprintf("User %s [%s]; Message %s for %s; %s",
                     $this->user->get_username(),
                     rcube_utils::remote_addr(),
+                    $headers['Message-ID'],
                     implode(', ', $mailto),
-                    !empty($response) ? join('; ', $response) : ''));
+                    !empty($response) ? implode('; ', $response) : ''));
             }
         }
 
@@ -1697,7 +1720,10 @@ class rcube
             $this->smtp->disconnect();
         }
 
-        $message->headers($headers, true);
+        // Add Bcc header back
+        if (!empty($headers['Bcc'])) {
+            $message->headers(array('Bcc' => $headers['Bcc']), true);
+        }
 
         return $sent;
     }
@@ -1714,6 +1740,11 @@ class rcube_dummy_plugin_api
 {
     /**
      * Triggers a plugin hook.
+     *
+     * @param string $hook Hook name
+     * @param array  $args Hook arguments
+     *
+     * @return array Hook arguments
      * @see rcube_plugin_api::exec_hook()
      */
     public function exec_hook($hook, $args = array())
