@@ -3,7 +3,8 @@
 /**
  +-----------------------------------------------------------------------+
  | This file is part of the Roundcube Webmail client                     |
- | Copyright (C) 2008-2014, The Roundcube Dev Team                       |
+ |                                                                       |
+ | Copyright (C) The Roundcube Dev Team                                  |
  |                                                                       |
  | Licensed under the GNU General Public License version 3 or            |
  | any later version with exceptions for skins & plugins.                |
@@ -23,12 +24,11 @@
  *
  * @package    Framework
  * @subpackage Storage
- * @author     Thomas Bruederli <roundcube@gmail.com>
  */
 class rcube_message
 {
     /**
-     * Instace of framework class.
+     * Instance of framework class.
      *
      * @var rcube
      */
@@ -54,6 +54,7 @@ class rcube_message
     public $folder;
     public $headers;
     public $sender;
+    public $context;
     public $parts        = array();
     public $mime_parts   = array();
     public $inline_parts = array();
@@ -78,11 +79,17 @@ class rcube_message
     function __construct($uid, $folder = null, $is_safe = false)
     {
         // decode combined UID-folder identifier
-        if (preg_match('/^\d+-.+/', $uid)) {
+        if (preg_match('/^[0-9.]+-.+/', $uid)) {
             list($uid, $folder) = explode('-', $uid, 2);
         }
 
+        if (preg_match('/^([0-9]+)\.([0-9.]+)$/', $uid, $matches)) {
+            $uid     = $matches[1];
+            $context = $matches[2];
+        }
+
         $this->uid     = $uid;
+        $this->context = $context;
         $this->app     = rcube::get_instance();
         $this->storage = $this->app->get_storage();
         $this->folder  = strlen($folder) ? $folder : $this->storage->get_folder();
@@ -97,10 +104,6 @@ class rcube_message
             return;
         }
 
-        $this->mime    = new rcube_mime($this->headers->charset);
-        $this->subject = $this->headers->get('subject');
-        list(, $this->sender) = each($this->mime->decode_address_list($this->headers->from, 1));
-
         $this->set_safe($is_safe || $_SESSION['safe_messages'][$this->folder.':'.$uid]);
         $this->opt = array(
             'safe'        => $this->is_safe,
@@ -116,10 +119,15 @@ class rcube_message
             $this->get_mime_numbers($this->headers->structure);
             $this->parse_structure($this->headers->structure);
         }
-        else {
+        else if ($this->context === null) {
             $this->body = $this->storage->get_body($uid);
         }
 
+        $this->mime    = new rcube_mime($this->headers->charset);
+        $this->subject = $this->headers->get('subject');
+        $from          = $this->mime->decode_address_list($this->headers->from, 1);
+        $this->sender  = current($from);
+
         // notify plugins and let them analyze this structured message object
         $this->app->plugins->exec_hook('message_load', array('object' => $this));
     }
@@ -128,7 +136,8 @@ class rcube_message
      * Return a (decoded) message header
      *
      * @param string $name Header name
-     * @param bool   $row  Don't mime-decode the value
+     * @param bool   $raw  Don't mime-decode the value
+     *
      * @return string Header value
      */
     public function get_header($name, $raw = false)
@@ -155,6 +164,7 @@ class rcube_message
      *
      * @param string $mime_id Part MIME-ID
      * @param mixed  $embed Mimetype class for parts to be embedded
+     *
      * @return string URL or false if part does not exist
      */
     public function get_part_url($mime_id, $embed = false)
@@ -169,7 +179,7 @@ class rcube_message
      * Get content of a specific part of this message
      *
      * @param string   $mime_id           Part MIME-ID
-     * @param resource $fp File           pointer to save the message part
+     * @param resource $fp                File pointer to save the message part
      * @param boolean  $skip_charset_conv Disables charset conversion
      * @param int      $max_bytes         Only read this number of bytes
      * @param boolean  $formatted         Enables formatting of text/* parts bodies
@@ -199,11 +209,11 @@ class rcube_message
     /**
      * Get content of a specific part of this message
      *
-     * @param string  $mime_id   Part ID
-     * @param boolean $formatted Enables formatting of text/* parts bodies
-     * @param int     $max_bytes Only return/read this number of bytes
-     * @param mixed   $mode      NULL to return a string, -1 to print body
-     *                           or file pointer to save the body into
+     * @param string $mime_id   Part ID
+     * @param bool   $formatted Enables formatting of text/* parts bodies
+     * @param int    $max_bytes Only return/read this number of bytes
+     * @param mixed  $mode      NULL to return a string, -1 to print body
+     *                          or file pointer to save the body into
      *
      * @return string|bool Part content or operation status
      */
@@ -243,7 +253,7 @@ class rcube_message
             if (is_resource($mode)) {
                 if ($body !== false) {
                     fwrite($mode, $body);
-                    rewind($mode);
+                    @rewind($mode);
                 }
 
                 return $body !== false;
@@ -268,7 +278,7 @@ class rcube_message
             !($mode && $formatted), $max_bytes, $mode && $formatted);
 
         if (is_resource($mode)) {
-            rewind($mode);
+            @rewind($mode);
             return $body !== false;
         }
 
@@ -338,6 +348,14 @@ class rcube_message
                     continue;
                 }
 
+                if (!$part->size) {
+                    continue;
+                }
+
+                if (!$this->check_context($part)) {
+                    continue;
+                }
+
                 $level = explode('.', $part->mime_id);
                 $depth = count($level);
                 $last  = '';
@@ -350,19 +368,22 @@ class rcube_message
                         return true;
                     }
 
-                    $parent    = $this->mime_parts[join('.', $level)];
-                    $max_delta = $depth - (1 + ($last == 'multipart/alternative' ? 1 : 0));
-                    $last      = $parent->mimetype;
+                    $parent = $this->mime_parts[implode('.', $level)];
 
-                    if (!preg_match('/^multipart\/(alternative|related|signed|encrypted|mixed)$/', $parent->mimetype)
-                        || ($parent->mimetype == 'multipart/mixed' && $parent_depth < $max_delta)) {
+                    if (!$this->check_context($parent)) {
+                        return true;
+                    }
+
+                    $max_delta = $depth - (1 + ($last == 'multipart/alternative' ? 1 : 0));
+                    $last      = $parent->real_mimetype ?: $parent->mimetype;
+
+                    if (!preg_match('/^multipart\/(alternative|related|signed|encrypted|mixed)$/', $last)
+                        || ($last == 'multipart/mixed' && $parent_depth < $max_delta)) {
                         continue 2;
                     }
                 }
 
-                if ($part->size) {
-                    return true;
-                }
+                return true;
             }
         }
 
@@ -389,6 +410,14 @@ class rcube_message
                     continue;
                 }
 
+                if (!$part->size) {
+                    continue;
+                }
+
+                if (!$this->check_context($part)) {
+                    continue;
+                }
+
                 $level = explode('.', $part->mime_id);
 
                 // Check if the part belongs to higher-level's alternative/related
@@ -397,15 +426,18 @@ class rcube_message
                         return true;
                     }
 
-                    $parent = $this->mime_parts[join('.', $level)];
+                    $parent = $this->mime_parts[implode('.', $level)];
+
+                    if (!$this->check_context($parent)) {
+                        return true;
+                    }
+
                     if ($parent->mimetype != 'multipart/alternative' && $parent->mimetype != 'multipart/related') {
                         continue 2;
                     }
                 }
 
-                if ($part->size) {
-                    return true;
-                }
+                return true;
             }
         }
 
@@ -463,6 +495,26 @@ class rcube_message
         }
     }
 
+    /**
+     * Return message parts in current context
+     */
+    public function mime_parts()
+    {
+        if ($this->context === null) {
+            return $this->mime_parts;
+        }
+
+        $parts = array();
+
+        foreach ($this->mime_parts as $part_id => $part) {
+            if ($this->check_context($part)) {
+                $parts[$part_id] = $part;
+            }
+        }
+
+        return $parts;
+    }
+
     /**
      * Checks if part of the message is an attachment (or part of it)
      *
@@ -527,6 +579,10 @@ class rcube_message
             if (!isset($structure->headers['subject']) && !isset($structure->headers['from'])) {
                 list($headers, ) = explode("\r\n\r\n", $this->get_part_body($structure->mime_id, false, 32768));
                 $structure->headers = rcube_mime::parse_headers($headers);
+
+                if ($this->context == $structure->mime_id) {
+                    $this->headers = rcube_message_header::from_array($structure->headers);
+                }
             }
         }
         else {
@@ -535,11 +591,12 @@ class rcube_message
 
         // show message headers
         if ($recursive && is_array($structure->headers) &&
-                (isset($structure->headers['subject']) || $structure->headers['from'] || $structure->headers['to'])) {
+            (isset($structure->headers['subject']) || $structure->headers['from'] || $structure->headers['to'])
+        ) {
             $c = new stdClass;
             $c->type = 'headers';
             $c->headers = $structure->headers;
-            $this->parts[] = $c;
+            $this->add_part($c);
         }
 
         // Allow plugins to handle message parts
@@ -561,25 +618,25 @@ class rcube_message
         if ($message_ctype_primary == 'text' && !$recursive) {
             // parts with unsupported type add to attachments list
             if (!in_array($message_ctype_secondary, array('plain', 'html', 'enriched'))) {
-                $this->attachments[] = $structure;
+                $this->add_part($structure, 'attachment');
                 return;
             }
 
             $structure->type = 'content';
-            $this->parts[] = $structure;
+            $this->add_part($structure);
 
             // Parse simple (plain text) message body
             if ($message_ctype_secondary == 'plain') {
                 foreach ((array)$this->uu_decode($structure) as $uupart) {
                     $this->mime_parts[$uupart->mime_id] = $uupart;
-                    $this->attachments[] = $uupart;
+                    $this->add_part($uupart, 'attachment');
                 }
             }
         }
         // the same for pgp signed messages
         else if ($mimetype == 'application/pgp' && !$recursive) {
             $structure->type = 'content';
-            $this->parts[] = $structure;
+            $this->add_part($structure);
         }
         // message contains (more than one!) alternative parts
         else if ($mimetype == 'multipart/alternative'
@@ -613,7 +670,7 @@ class rcube_message
                     $enriched_part = $p;
                 else {
                     // add unsupported/unrecognized parts to attachments list
-                    $this->attachments[] = $sub_part;
+                    $this->add_part($sub_part, 'attachment');
                 }
             }
 
@@ -624,8 +681,9 @@ class rcube_message
                 $this->parse_alternative = false;
 
                 // if plain part was found, we should unset it if html is preferred
-                if ($this->opt['prefer_html'] && count($this->parts))
+                if ($this->opt['prefer_html'] && count($this->parts)) {
                     $plain_part = null;
+                }
             }
 
             // choose html/plain part to print
@@ -642,7 +700,15 @@ class rcube_message
             // add the right message body
             if (is_object($print_part)) {
                 $print_part->type = 'content';
-                $this->parts[] = $print_part;
+
+                // Allow plugins to handle also this part
+                $plugin = $this->app->plugins->exec_hook('message_part_structure',
+                    array('object' => $this, 'structure' => $print_part,
+                        'mimetype' => $print_part->mimetype, 'recursive' => true));
+
+                if (!$plugin['abort']) {
+                    $this->add_part($print_part);
+                }
             }
             // show plaintext warning
             else if ($html_part !== null && empty($this->parts)) {
@@ -653,7 +719,7 @@ class rcube_message
                 $c->mimetype        = 'text/plain';
                 $c->realtype        = 'text/html';
 
-                $this->parts[] = $c;
+                $this->add_part($c);
             }
         }
         // this is an ecrypted message -> create a plaintext body with the according message
@@ -666,14 +732,14 @@ class rcube_message
             $p->realtype        = 'multipart/encrypted';
             $p->mime_id         = $structure->mime_id;
 
-            $this->parts[] = $p;
+            $this->add_part($p);
 
             // add encrypted payload part as attachment
             if (is_array($structure->parts)) {
                 for ($i=0; $i < count($structure->parts); $i++) {
                     $subpart = $structure->parts[$i];
                     if ($subpart->mimetype == 'application/octet-stream' || !empty($subpart->filename)) {
-                        $this->attachments[] = $subpart;
+                        $this->add_part($subpart, 'attachment');
                     }
                 }
             }
@@ -688,10 +754,10 @@ class rcube_message
             $p->realtype        = 'application/pkcs7-mime';
             $p->mime_id         = $structure->mime_id;
 
-            $this->parts[] = $p;
+            $this->add_part($p);
 
             if (!empty($structure->filename)) {
-                $this->attachments[] = $structure;
+                $this->add_part($structure, 'attachment');
             }
         }
         // message contains multiple parts
@@ -705,12 +771,12 @@ class rcube_message
 
                 // multipart/alternative or message/rfc822
                 if ($primary_type == 'multipart' || $part_mimetype == 'message/rfc822') {
-                    $this->parse_structure($mail_part, true);
-
-                    // list message/rfc822 as attachment as well (mostly .eml)
-                    if ($primary_type == 'message' && !empty($mail_part->filename)) {
-                        $this->attachments[] = $mail_part;
+                    // list message/rfc822 as attachment as well
+                    if ($part_mimetype == 'message/rfc822') {
+                        $this->add_part($mail_part, 'attachment');
                     }
+
+                    $this->parse_structure($mail_part, true);
                 }
                 // part text/[plain|html] or delivery status
                 else if ((($part_mimetype == 'text/plain' || $part_mimetype == 'text/html') && $mail_part->disposition != 'attachment') ||
@@ -738,12 +804,12 @@ class rcube_message
                         ($secondary_type == 'plain' && !$this->opt['prefer_html'])
                     ) {
                         $mail_part->type = 'content';
-                        $this->parts[] = $mail_part;
+                        $this->add_part($mail_part);
                     }
 
                     // list as attachment as well
                     if (!empty($mail_part->filename)) {
-                        $this->attachments[] = $mail_part;
+                        $this->add_part($mail_part, 'attachment');
                     }
                 }
                 // ignore "virtual" protocol parts
@@ -755,13 +821,13 @@ class rcube_message
                     $tnef_parts = (array) $this->tnef_decode($mail_part);
                     foreach ($tnef_parts as $tpart) {
                         $this->mime_parts[$tpart->mime_id] = $tpart;
-                        $this->attachments[] = $tpart;
+                        $this->add_part($tpart, 'attachment');
                     }
 
                     // add winmail.dat to the list if it's content is unknown
                     if (empty($tnef_parts) && !empty($mail_part->filename)) {
                         $this->mime_parts[$mail_part->mime_id] = $mail_part;
-                        $this->attachments[] = $mail_part;
+                        $this->add_part($mail_part, 'attachment');
                     }
                 }
                 // part is a file/attachment
@@ -771,33 +837,37 @@ class rcube_message
                         (empty($mail_part->disposition) || preg_match('/^[a-z0-9!#$&.+^_-]+$/i', $mail_part->disposition)))
                 ) {
                     // skip apple resource forks
-                    if ($message_ctype_secondary == 'appledouble' && $secondary_type == 'applefile')
+                    if ($message_ctype_secondary == 'appledouble' && $secondary_type == 'applefile') {
                         continue;
+                    }
+
+                    if ($mail_part->headers['content-id']) {
+                        $mail_part->content_id = preg_replace(array('/^</', '/>$/'), '', $mail_part->headers['content-id']);
+                    }
+
+                    if ($mail_part->headers['content-location']) {
+                        $mail_part->content_location = $mail_part->headers['content-base'] . $mail_part->headers['content-location'];
+                    }
 
                     // part belongs to a related message and is linked
-                    if (preg_match('/^multipart\/(related|relative)/', $mimetype)
-                        && ($mail_part->headers['content-id'] || $mail_part->headers['content-location'])
+                    // Note: mixed is not supposed to contain inline images, but we've found such examples (#5905)
+                    if (preg_match('/^multipart\/(related|relative|mixed)/', $mimetype)
+                        && ($mail_part->content_id || $mail_part->content_location)
                     ) {
-                        if ($mail_part->headers['content-id'])
-                            $mail_part->content_id = preg_replace(array('/^</', '/>$/'), '', $mail_part->headers['content-id']);
-                        if ($mail_part->headers['content-location'])
-                            $mail_part->content_location = $mail_part->headers['content-base'] . $mail_part->headers['content-location'];
-
-                        $this->inline_parts[] = $mail_part;
+                        $this->add_part($mail_part, 'inline');
                     }
-                    // regular attachment with valid content type
-                    // (content-type name regexp according to RFC4288.4.2)
-                    else if (preg_match('/^[a-z0-9!#$&.+^_-]+\/[a-z0-9!#$&.+^_-]+$/i', $part_mimetype)) {
-                        $this->attachments[] = $mail_part;
-                    }
-                    // attachment with invalid content type
-                    // replace malformed content type with application/octet-stream (#1487767)
-                    else if ($mail_part->filename) {
-                        $mail_part->ctype_primary   = 'application';
-                        $mail_part->ctype_secondary = 'octet-stream';
-                        $mail_part->mimetype        = 'application/octet-stream';
 
-                        $this->attachments[] = $mail_part;
+                    // Any non-inline attachment
+                    if (!preg_match('/^inline/i', $mail_part->disposition) || empty($mail_part->headers['content-id'])) {
+                        // Content-Type name regexp according to RFC4288.4.2
+                        if (!preg_match('/^[a-z0-9!#$&.+^_-]+\/[a-z0-9!#$&.+^_-]+$/i', $part_mimetype)) {
+                            // replace malformed content type with application/octet-stream (#1487767)
+                            $mail_part->ctype_primary   = 'application';
+                            $mail_part->ctype_secondary = 'octet-stream';
+                            $mail_part->mimetype        = 'application/octet-stream';
+                        }
+
+                        $this->add_part($mail_part, 'attachment');
                     }
                 }
                 // calendar part not marked as attachment (#1490325)
@@ -806,19 +876,21 @@ class rcube_message
                         $mail_part->filename = 'calendar.ics';
                     }
 
-                    $this->attachments[] = $mail_part;
+                    $this->add_part($mail_part, 'attachment');
                 }
             }
 
-            // if this was a related part try to resolve references
-            if (preg_match('/^multipart\/(related|relative)/', $mimetype) && sizeof($this->inline_parts)) {
+            // if this is a related part try to resolve references
+            // Note: mixed is not supposed to contain inline images, but we've found such examples (#5905)
+            if (preg_match('/^multipart\/(related|relative|mixed)/', $mimetype) && count($this->inline_parts)) {
                 $a_replaces = array();
                 $img_regexp = '/^image\/(gif|jpe?g|png|tiff|bmp|svg)/';
 
                 foreach ($this->inline_parts as $inline_object) {
                     $part_url = $this->get_part_url($inline_object->mime_id, $inline_object->ctype_primary);
-                    if (isset($inline_object->content_id))
+                    if (isset($inline_object->content_id)) {
                         $a_replaces['cid:'.$inline_object->content_id] = $part_url;
+                    }
                     if ($inline_object->content_location) {
                         $a_replaces[$inline_object->content_location] = $part_url;
                     }
@@ -828,13 +900,13 @@ class rcube_message
                         // In this case multipart/related message has only one text part
                         // We'll add all such attachments to the attachments list
                         if (!isset($this->got_html_part)) {
-                            $this->attachments[] = $inline_object;
+                            $this->add_part($inline_object, 'attachment');
                         }
                         // MS Outlook sometimes also adds non-image attachments as related
                         // We'll add all such attachments to the attachments list
                         // Warning: some browsers support pdf in <img/>
                         else if (!preg_match($img_regexp, $inline_object->mimetype)) {
-                            $this->attachments[] = $inline_object;
+                            $this->add_part($inline_object, 'attachment');
                         }
                         // @TODO: we should fetch HTML body and find attachment's content-id
                         // to handle also image attachments without reference in the body
@@ -845,41 +917,73 @@ class rcube_message
                 // add replace array to each content part
                 // (will be applied later when part body is available)
                 foreach ($this->parts as $i => $part) {
-                    if ($part->type == 'content')
+                    if ($part->type == 'content') {
                         $this->parts[$i]->replaces = $a_replaces;
+                    }
                 }
             }
         }
         // message is a single part non-text
-        else if ($structure->filename) {
-            $this->attachments[] = $structure;
-        }
-        // message is a single part non-text (without filename)
-        else if (preg_match('/application\//i', $mimetype)) {
-            $this->attachments[] = $structure;
+        else if ($structure->filename || preg_match('/^application\//i', $mimetype)) {
+            $this->add_part($structure, 'attachment');
         }
     }
 
     /**
-     * Fill aflat array with references to all parts, indexed by part numbers
+     * Fill a flat array with references to all parts, indexed by part numbers
      *
      * @param rcube_message_part $part Message body structure
      */
     private function get_mime_numbers(&$part)
     {
-        if (strlen($part->mime_id))
+        if (strlen($part->mime_id)) {
             $this->mime_parts[$part->mime_id] = &$part;
+        }
 
-        if (is_array($part->parts))
-            for ($i=0; $i<count($part->parts); $i++)
+        if (is_array($part->parts)) {
+            for ($i=0; $i<count($part->parts); $i++) {
                 $this->get_mime_numbers($part->parts[$i]);
+            }
+        }
+    }
+
+    /**
+     * Add a part to object parts array(s) (with context check)
+     *
+     * @param rcube_message_part $part Message part
+     * @param string             $type Part type (inline/attachment)
+     */
+    private function add_part($part, $type = null)
+    {
+        if ($this->check_context($part)) {
+            // It may happen that we add the same part to the array many times
+            // use part ID index to prevent from duplicates
+            switch ($type) {
+                case 'inline': $this->inline_parts[(string) $part->mime_id] = $part; break;
+                case 'attachment': $this->attachments[(string) $part->mime_id] = $part; break;
+                default: $this->parts[] = $part; break;
+            }
+        }
+    }
+
+    /**
+     * Check if specified part belongs to the current context
+     *
+     * @param rcube_message_part $part Message part
+     *
+     * @return bool True if the part belongs to the current context, False otherwise
+     */
+    private function check_context($part)
+    {
+        return $this->context === null || strpos($part->mime_id, $this->context . '.') === 0;
     }
 
     /**
      * Decode a Microsoft Outlook TNEF part (winmail.dat)
      *
      * @param rcube_message_part $part Message part to decode
-     * @return array
+     *
+     * @return rcube_message_part[] List of message parts extracted from TNEF
      */
     function tnef_decode(&$part)
     {
@@ -891,7 +995,7 @@ class rcube_message
 
         unset($body);
 
-        foreach ($tnef_arr as $pid => $winatt) {
+        foreach ($tnef_arr['attachments'] as $pid => $winatt) {
             $tpart = new rcube_message_part;
 
             $tpart->filename        = $this->fix_attachment_name(trim($winatt['name']), $part);
@@ -914,7 +1018,8 @@ class rcube_message
      * Parse message body for UUencoded attachments bodies
      *
      * @param rcube_message_part $part Message part to decode
-     * @return array
+     *
+     * @return rcube_message_part[] List of message parts extracted from the file
      */
     function uu_decode(&$part)
     {
@@ -967,7 +1072,12 @@ class rcube_message
     }
 
     /**
-     * Fix attachment name encoding if needed/possible
+     * Fix attachment name encoding if needed and possible
+     *
+     * @param string             $name Attachment name
+     * @param rcube_message_part $part Message part
+     *
+     * @return string Fixed attachment name
      */
     protected function fix_attachment_name($name, $part)
     {
@@ -989,8 +1099,8 @@ class rcube_message
             // check parents' charset
             $items = explode('.', $part->mime_id);
             for ($i = count($items)-1; $i > 0; $i--) {
-                $last   = array_pop($items);
-                $parent = $this->mime_parts[join('.', $items)];
+                array_pop($items);
+                $parent = $this->mime_parts[implode('.', $items)];
 
                 if ($parent && $parent->charset) {
                     $charsets[] = $parent->charset;
